%{
/* C declarations */

#define YYDEBUG 1
#define YYERROR_VERBOSE 1

%}
/* BISON declarations */

/* Expect 2 shift/reduce conflicts. */
%expect 2
/*   IF rule1 THEN IF rule2 THEN st1 ELSE st2
 * can be translated as:
 *   - st2 if not rule1
 *   - st2 if not rule2
 */
/*   IF rule THEN command1: command2
 * can be translated as:
 *   - command2 regardless
 *   - command2 only if command1 too
 */




/* Operator precedence - least precedence group */
%left IMP EQV   /* 1101, 1001 */
%left XOR OR    /* 0110, 0111 */
%left AND       /* 0001 */

%right NOT  /* 0=1 1=0 */

/* Operator precedence - middle precedence group */
%left '<' '>' '=' LESSOREQUAL GREATEROREQUAL NOTEQUAL

/* Operator precedence - top precedence group */
%left '+' '-'
%left MOD
%left '\\'
%left '*' '/'

%right NEGATIONPRIORITY
%right '^'

%token CRLF COLONCHAR

%token IF THEN ELSE ELSEIF ENDIF

%token SELECT CASE ENDSELECT IS

%token WEND DO LOOP WHILE EXITDO
%token DOWHILE DOUNTIL LOOPWHILE LOOPUNTIL

%token SUB ENDSUB EXITSUB DECLARESUB
%token FUNCTION ENDFUNCTION EXITFUNCTION DECLAREFUNCTION
%token CALL

%token FOR TO STEP NEXT EXITFOR
%token GOTO GOSUB RETURN RESTORE
%token END
%token LET

%token NUMBER STRING
%token IDENTIFIER FUNCTIONNAME SUBNAME CONSTNAME

%token PRINT USING FILENUM LPRINT OPEN
%token OUTPUT INPUT APPEND RANDOM BINARY WRITE ACCESS LOCK CLOSE

%token ON PLAY STOP OFF POKE SYSTEM LOCATE ERROR RESUME
%token DEFSEG

%token INKEY_S CSRLIN SPACE_S POS LEN FREEFILE

%token DATA READ

%token INTEGER SINGLE DOUBLE STRING LONG AS
%token TYPE ENDTYPE
%token DIM REDIM SHARED CONST

%token UNKNOWN_CHARACTER

%%
/* Grammar rules */

program:
		definitions0
;

definitions0:
		definitions0 definition1
	|	/* empty */
;

definition1:
		sub
	|	function
	|	declaresub
	|	declarefunction
	|	typedef
	|	constdef
	|	longstatement1 separator1
	|	anylabeldef
	|	separator1
;

sub:
		SUB autosubname sub_params_decl separator1
		    longstatements0 ENDSUB CRLF
;
function:
		FUNCTION autofunctionname sub_params_decl separator1
		         longstatements0 ENDFUNCTION CRLF
;
declaresub:
		DECLARESUB autosubname sub_params_decl CRLF
;
declarefunction:
		DECLAREFUNCTION autofunctionname sub_params_decl CRLF
;

typedef:
		TYPE autousertype CRLF typeelements2 ENDTYPE CRLF
;


constdef:
		CONST autoconstname '=' expr CRLF
;

/* This does not terminate with anything */
shortstatement1:
		simplestatement
	|	shortif
	|	shortwhile
	|	shortdoloop
	|	shortfornext
	|	shortselect
;
shortstatements0: /* Used where short statements are required */
		shortstatements0 shortstatement1 COLONCHAR
	|	shortstatements0 COLONCHAR
	|	/* empty */
;
shortstatements2: /* Used in IF */
		shortstatements2 COLONCHAR shortstatement1
	|	shortstatement1
;

shortif:
	/* No shortifblock0 rules here. Would cause reduce/reduce conflicts. */
		IF boolexpr THEN shortifblock2 ELSE shortifblock2
	|	IF boolexpr THEN               ELSE shortifblock2
	|	IF boolexpr THEN shortifblock2 ELSE
	|	IF boolexpr THEN               ELSE
	|	IF boolexpr THEN shortifblock2
;
shortifblock2:
		NUMBER /* gotolabel */
	|	shortstatements2
;

shortdoloop:
		DO COLONCHAR shortstatements0 LOOP
	|	DOWHILE boolexpr COLONCHAR shortstatements0 LOOP
	|	DOUNTIL boolexpr COLONCHAR shortstatements0 LOOP
	|	DO COLONCHAR shortstatements0 LOOPWHILE boolexpr
	|	DO COLONCHAR shortstatements0 LOOPUNTIL boolexpr
;
shortwhile:
		WHILE boolexpr COLONCHAR shortstatements0 WEND
;
shortfornext:
		FOR target '=' numexpr TO numexpr forstep COLONCHAR shortstatements0 NEXT
;
shortselect:
		SELECT CASE expr colonchar2 shortcaselist ENDSELECT
;
shortcaselist:
		shortcaselist CASE casevalues2 COLONCHAR shortstatements0
	|	/* empty */
;
colonchar2:
		colonchar2 COLONCHAR
	|	COLONCHAR
;

longstatement1:
		longstructure
	|	simplestatement
	|	shortif
;
longstatements0:
		longstatements0 longstatement1 separator1
	|	longstatements0 anylabeldef
	|	longstatements0 separator1
	|	/* empty */
;

longstructure:
		longif
	|	longwhile
	|	longdoloop
	|	longfornext
	|	longselect
;
longif:
		IF boolexpr THEN CRLF longstatements0 longelselist ENDIF
;
longelselist:
		ELSEIF boolexpr THEN longstatements0 longelselist
	|	ELSE longstatements0
	|	/* empty */
;
longdoloop:
		DO separator1 longstatements0 LOOP
	|	DOWHILE boolexpr separator1 longstatements0 LOOP
	|	DOUNTIL boolexpr separator1 longstatements0 LOOP
	|	DO separator1 longstatements0 LOOPWHILE boolexpr
	|	DO separator1 longstatements0 LOOPUNTIL boolexpr
;
longwhile:
		WHILE boolexpr separator1 longstatements0 WEND
;
longfornext:
		FOR target '=' numexpr TO numexpr forstep separator1 longstatements0 NEXT
;
longselect:
		SELECT CASE expr separators2 longcaselist ENDSELECT
;
longcaselist:
		longcaselist CASE casevalues2 separator1 longstatements0
	|	/* empty */
;


forstep:
		STEP numexpr
	|	/* empty */
;
casevalue1:
		ELSE
	|	IS '>' expr
	|	IS '<' expr
	|	IS '=' expr
	|	IS LESSOREQUAL expr
	|	IS GREATEROREQUAL expr
	|	IS NOTEQUAL expr
	|	expr TO expr
	|	expr
;
casevalues2:
		casevalues2 ',' casevalue1
	|	casevalue1
;

separator1:
		COLONCHAR
	|	CRLF
;
/* unused
separators0:
		separators2
	|	-empty-
;
*/
separators2:
		separators2 separator1
	|	separator1
;

anylabeldef:
		NUMBER
	|	autolabelname COLONCHAR
;
anylabel1:
		autolabelname
	|	NUMBER
;
anylabel2:
		anylabel2 ',' anylabel1
	|	anylabel1
;


/* Something that is not a control structure */
/* This does not terminate with anything */
simplestatement:
		LET assignment
	|	assignment
	|	RETURN anylabel1
	|	RETURN
	|	EXITDO
	|	EXITFOR
	|	EXITSUB
	|	EXITFUNCTION
	|	subcall
	|	dim
	|	GOTO anylabel1
	|	GOSUB anylabel1
	|	RESTORE anylabel1
	|	RESTORE
	|	ON PLAY '(' expr ')' GOSUB anylabel1
	|	ON numexpr GOSUB anylabel2
	|	ON ERROR GOTO anylabel1
	|	RESUME anylabel1
	|	RESUME NEXT
	|	RESUME
	|	PRINT FILENUM ',' printspec
	|	PRINT printspec
	|	LPRINT printspec
	|	PLAY STOP
	|	PLAY OFF
	|	PLAY ON
	|	PLAY stringexpr
	|	POKE numexpr ',' numexpr
	|	DATA dataparams0
	|	SYSTEM stringexpr
	|	READ readparams2
	|	SYSTEM
	|	END
	|	locate
	|	open
	|	CLOSE closeparams0
;

dataparams0:
		dataparams0 ',' STRING
	|	dataparams0 ','
	|	STRING
	|	/* empty */
;
readparams2:
		readparams2 ',' target
	|	target
;

closeparams0:
		closeparams0 ',' autofilenum
	|	autofilenum
	|	/* empty */
;

locate:
		LOCATE optional_numexpr ',' optional_numexpr ',' optional_numexpr ',' numexpr ',' numexpr
	                 /* row, col, cursor_visible, cursor_start, cursor_end */
	|	LOCATE optional_numexpr ',' optional_numexpr ',' optional_numexpr
	                 /* row, col, cursor_visible */
	|	LOCATE optional_numexpr ',' optional_numexpr
	                 /* row, col */
	|	LOCATE optional_numexpr
	                 /* row */
;

open:		OPEN stringexpr /* filename */
                     open_formode open_accessmode open_lockmode
                     AS autofilenum
                     open_lenmode
        |	OPEN stringexpr /* mode */
                     ',' autofilenum
                     ',' stringexpr /* filename */
                     open_reclenmode
;

open_formode:
		FOR OUTPUT
	|	FOR INPUT
	|	FOR APPEND
	|	FOR RANDOM
	|	FOR BINARY
	|	/* empty - RANDOM assumed */
;
open_accessmode:
		ACCESS READ WRITE
	|	ACCESS READ
	|	ACCESS WRITE
	|	/* empty */
;
open_lockmode:
		SHARED
	|	LOCK READ WRITE
	|	LOCK READ
	|	LOCK WRITE
	|	/* empty */
;
open_lenmode:
		LEN '=' numexpr
	|	/* empty */
;
open_reclenmode:
		',' numexpr
	|	/* empty */
;

assignment:
		target '=' expr
;

subcall:
		SUBNAME subcallparams0
	|	CALL autosubname '(' subcallparams0 ')'
	|	CALL autosubname
	|	IDENTIFIER subcallparamlist_no_parenthesis
		/* Calling an undefined SUB - autodefine! */
		/* No autosubname here - otherwise it conflicts with the first rule! */
	/*|	autosubname
	   - can't be done: It confuses with label
	 */
;

autosubname:
		SUBNAME
	|	IDENTIFIER
;
autofunctionname:
		FUNCTIONNAME
	|	IDENTIFIER
;
autolabelname:
		IDENTIFIER
;
autousertype:
		IDENTIFIER
;
autovarname:
		IDENTIFIER
;
autoconstname:
		IDENTIFIER
;
autofilenum:
		FILENUM
	|	numexpr
;

subcallparamlist_no_parenthesis:
		expr_no_parenthesis
	|	expr_no_parenthesis ',' subcallparams2
;

subcallparams0:
		subcallparams2
	|	/* empty */
;
subcallparams2:
		subcallparams2 ',' expr
	|	expr
;


sub_params_decl:
		'(' paramdeclarations2 ')'
	|	'(' ')'
	|	/* empty */
;
paramdeclaration1:
		autovarname AS type
	|	autovarname
;
paramdeclarations2:
		paramdeclarations2 ',' paramdeclaration1
	|	paramdeclaration1
;


typeelements2:
		typeelements2 typeelement1
	|	typeelement1
;
typeelement1:
		typeelementdef CRLF
	|	CRLF /* empty line is ok in typedef */
;
typeelementdef:
		typeelementdef ',' autovarname AS type /* extension */
	|	autovarname AS type
;




printspec:
		USING stringexpr ';' printparams0
	|	printparams0
;
printparams0:
		printparams0 ';' expr
	|	printparams0 ';'
	|	printparams0 ',' expr
	|	printparams0 ','
	|	expr
	|	/* empty */
;

dim:
		DIM SHARED dimdefs2
	|	DIM dimdefs2
	|	REDIM redimdefs2
;
dimdefs2:
		dimdefs2 ',' dimdef1
	|	dimdef1
;
dimdef1:
		autovarname '(' dimindexdefs2 ')' AS type
	|	autovarname '(' dimindexdefs2 ')'
	|	autovarname AS type
	|	autovarname
;
dimindexdefs2:
		dimindexdefs2 ',' dimindexdef1
	|	dimindexdef1
;
dimindexdef1:
		numexpr TO numexpr
	|	numexpr
;
redimdefs2:
		redimdefs2 ',' redimdef1
	|	redimdef1
;
redimdef1:
		autovarname '(' dimindexdefs2 ')' AS type
	|	autovarname '(' dimindexdefs2 ')'
;

expr:
		expr IMP expr
	|	expr EQV expr
	|	expr XOR expr
	|	expr OR expr
	|	expr AND expr
	|	NOT expr
	
	|	expr '>' expr
	|	expr '<' expr
	|	expr '=' expr
	|	expr LESSOREQUAL expr
	|	expr GREATEROREQUAL expr
	|	expr NOTEQUAL expr

	|	expr '+' expr
	|	expr '-' expr
	|	expr MOD expr
	|	expr '\\' expr
	|	expr '/' expr
	|	expr '*' expr
	|	expr '^' expr
	
	|       '-' expr %prec NEGATIONPRIORITY
	
	|	'(' expr ')'
	
	|	simple_expr
;

expr_no_parenthesis:
		expr_no_parenthesis IMP expr
	|	expr_no_parenthesis EQV expr
	|	expr_no_parenthesis XOR expr
	|	expr_no_parenthesis OR expr
	|	expr_no_parenthesis AND expr
	|	NOT expr
	
	|	expr_no_parenthesis '>' expr
	|	expr_no_parenthesis '<' expr
	|	expr_no_parenthesis '=' expr
	|	expr_no_parenthesis LESSOREQUAL expr
	|	expr_no_parenthesis GREATEROREQUAL expr
	|	expr_no_parenthesis NOTEQUAL expr

	|	expr_no_parenthesis '+' expr
	|	expr_no_parenthesis '-' expr
	|	expr_no_parenthesis MOD expr
	|	expr_no_parenthesis '\\' expr
	|	expr_no_parenthesis '/' expr
	|	expr_no_parenthesis '*' expr
	|	expr_no_parenthesis '^'expr
	
	|       '-' expr %prec NEGATIONPRIORITY
	
	|	simple_expr
;

numexpr:
		expr	/* FIXME - check expression type */
;
boolexpr:
		expr	/* FIXME - check expression type */
;
stringexpr:
		expr	/* FIXME - check expression type */
;

simple_expr:
		var
	|	functioncall
	|	builtinfunc
	|	NUMBER
	|	STRING
	|	CONSTNAME
;

optional_numexpr:
		numexpr
	|	/* empty - assume "-1" */
;

functioncall:
		FUNCTIONNAME '(' subcallparams2 ')'
		/* autofunctionname not ok here - confuses with array index */
	|	FUNCTIONNAME
;

indexlist:
		indexlist ',' numexpr
	|	numexpr
;

varname:
		autovarname '(' indexlist ')'
	|	autovarname
;
var:
		var '.' varname
	|	varname
;

target:
		var
;

type:
		INTEGER
	|	SINGLE
	|	DOUBLE
	|	STRING
	|	LONG
	|	autousertype
;

builtinfunc:
		INKEY_S
	|	CSRLIN
	|	FREEFILE
	|	SPACE_S '(' numexpr ')'
	|	POS '(' numexpr ')'
	|	LEN '(' stringexpr ')'
;

%%
/* Additional C code */


int main(void)
{
    /*setbuf(stdin,NULL);*/
    yydebug=1;
    yyparse();
    return 0;
}
